import os
import time
import json
import re
import inspect
from dataclasses import dataclass
from typing import Optional, Union, List, Dict


@dataclass
class Cache:
    """
    Declarative configuration for page caching.
    
    Usage in index.py:
        from casp.cache_handler import Cache
        Cache(ttl=600, enabled=True)
    """
    ttl: int = 600
    enabled: bool = True

    def __post_init__(self):
        # Auto-register to module globals (Magic Registration)
        try:
            # Stack Depth:
            # 0. __post_init__ (this function)
            # 1. __init__ (generated by dataclass)
            # 2. The module calling Cache(...) (e.g. index.py)
            frame = inspect.currentframe()
            if frame and frame.f_back and frame.f_back.f_back:
                caller_frame = frame.f_back.f_back
                
                # Only auto-register if called at module level
                if caller_frame.f_code.co_name == '<module>':
                    caller_frame.f_globals['cache_settings'] = self
        except Exception:
            pass


class CacheHandler:
    CACHE_DIR = os.path.join(os.getcwd(), 'caches')
    MANIFEST_FILE = os.path.join(CACHE_DIR, 'cache_manifest.json')

    @classmethod
    def ensure_cache_dir(cls):
        """Ensure cache dir and manifest file exist."""
        if not os.path.exists(cls.CACHE_DIR):
            os.makedirs(cls.CACHE_DIR, exist_ok=True)

        if not os.path.exists(cls.MANIFEST_FILE):
            cls._write_manifest({})

    @classmethod
    def _read_manifest(cls) -> Dict:
        if not os.path.exists(cls.MANIFEST_FILE):
            return {}
        try:
            with open(cls.MANIFEST_FILE, 'r', encoding='utf-8') as f:
                return json.load(f)
        except (json.JSONDecodeError, OSError):
            return {}

    @classmethod
    def _write_manifest(cls, data: Dict):
        try:
            with open(cls.MANIFEST_FILE, 'w', encoding='utf-8') as f:
                json.dump(data, f, indent=2)
        except OSError as e:
            print(f"[Cache] Error writing manifest: {e}")

    @staticmethod
    def get_filename(uri: str) -> str:
        clean_uri = uri.split('?')[0]
        if clean_uri == '/' or clean_uri == '':
            filename = 'index'
        else:
            filename = re.sub(r'[^a-zA-Z0-9\-_]', '_', clean_uri).lower()
            filename = filename.strip('_')
        return f"{filename}.html"

    @classmethod
    def serve_cache(cls, uri: str, default_ttl: int = 600) -> Optional[str]:
        if not uri:
            return None

        cls.ensure_cache_dir()
        manifest = cls._read_manifest()

        if uri not in manifest:
            return None

        entry = manifest[uri]
        html_file = os.path.join(cls.CACHE_DIR, entry['file'])

        if not os.path.exists(html_file):
            cls.invalidate_by_uri(uri)
            return None

        saved_ttl = entry.get('ttl', default_ttl)
        created_at = entry.get('created_at', 0)
        active_ttl = saved_ttl if saved_ttl > 0 else default_ttl
        age = time.time() - created_at

        if age > active_ttl:
            cls.invalidate_by_uri(uri)
            return None

        try:
            with open(html_file, 'r', encoding='utf-8') as f:
                content = f.read()
            return content
        except Exception:
            return None

    @classmethod
    def save_cache(cls, uri: str, content: str, ttl: int = 0):
        if not uri:
            return

        cls.ensure_cache_dir()
        filename = cls.get_filename(uri)
        html_path = os.path.join(cls.CACHE_DIR, filename)

        manifest = cls._read_manifest()
        manifest[uri] = {
            "file": filename,
            "ttl": ttl,
            "created_at": time.time()
        }
        cls._write_manifest(manifest)

        try:
            with open(html_path, 'w', encoding='utf-8') as f:
                f.write(content)
            print(f"[Cache] Saved: {uri} (TTL: {ttl}s)")
        except Exception as e:
            print(f"[Cache] Error saving HTML: {e}")

    @classmethod
    def invalidate_by_uri(cls, uris: Union[str, List[str]]):
        if isinstance(uris, str):
            uris = [uris]

        cls.ensure_cache_dir()
        manifest = cls._read_manifest()
        updated = False

        for uri in uris:
            normalized_uri = '/' + uri.lstrip('/')
            if normalized_uri in manifest:
                entry = manifest[normalized_uri]
                file_path = os.path.join(cls.CACHE_DIR, entry['file'])
                if os.path.exists(file_path):
                    try:
                        os.remove(file_path)
                    except OSError:
                        pass
                del manifest[normalized_uri]
                updated = True
                print(f"[Cache] Invalidated: {normalized_uri}")

        if updated:
            cls._write_manifest(manifest)

    @classmethod
    def reset_cache(cls, uri: Optional[str] = None):
        if uri:
            cls.invalidate_by_uri(uri)
            return
        if os.path.exists(cls.CACHE_DIR):
            for f in os.listdir(cls.CACHE_DIR):
                file_path = os.path.join(cls.CACHE_DIR, f)
                try:
                    os.remove(file_path)
                except OSError:
                    pass
        cls.ensure_cache_dir()
        cls._write_manifest({})


from __future__ import annotations
from typing import Any
from tailwind_merge import TailwindMerge

_twm = TailwindMerge()


def tw_merge(*values: Any) -> str:
    """
    Merge Tailwind class strings (and optionally lists/tuples/sets) into a single
    conflict-resolved class string using tailwind-merge.
    """
    parts: list[str] = []

    for v in values:
        if not v:
            continue

        if isinstance(v, (list, tuple, set)):
            chunk = " ".join(str(x) for x in v if x)
        else:
            chunk = str(v).strip()

        if chunk:
            parts.append(chunk)

    if not parts:
        return ""

    # TailwindMerge.merge accepts multiple class strings. :contentReference[oaicite:2]{index=2}
    return _twm.merge(*parts)



from __future__ import annotations
from typing import Callable
import os
import json
import hashlib
import importlib.util
import inspect
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple
from jinja2 import Environment, BaseLoader, select_autoescape
from markupsafe import Markup
import contextvars

_runtime_metadata = contextvars.ContextVar("casp_metadata", default=None)

class LayoutEngine:
    """
    Single source of truth for:
      - Jinja environment (Caspian delimiters)
      - Template loading (file + relative-to-py)
      - Rendering (string + page + layout)
      - Nested layout system (Next.js style layout discovery and wrapping)

    Canonical public entrypoints:
      - render_page(__file__, context)   -> index.html
      - render_layout(__file__, context) -> layout.html

    No load_page/load_layout helpers exist in this module by design.
    """

    def __init__(self) -> None:
        self.env = Environment(
            loader=BaseLoader(),
            autoescape=select_autoescape(["html", "xml"]),
            variable_start_string="[[",
            variable_end_string="]]",
            block_start_string="[%",  # Caspian blocks
            block_end_string="%]",
        )
        self.env.filters["json"] = self._to_safe_json
        self.env.filters["dump"] = self._to_safe_json

        # Legacy alias used elsewhere
        self.string_env = self.env

    # ---------------------------------------------------------------------
    # Jinja helpers
    # ---------------------------------------------------------------------

    @staticmethod
    def _to_safe_json(value: Any) -> Markup:
        """Safely serialize to JSON for HTML/JS embedding."""
        s = json.dumps(value, ensure_ascii=False, default=str)
        escaped = (
            s.replace("</", "<\\/")
            .replace("<!--", "<\\!--")
            .replace("\u2028", "\\u2028")
            .replace("\u2029", "\\u2029")
        )
        return Markup(escaped)

    # ---------------------------------------------------------------------
    # Core template loading & rendering
    # ---------------------------------------------------------------------

    def load_template_file(self, file_path: str) -> str:
        """Load template file directly by absolute/relative path."""
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"Template not found: {file_path}")
        with open(file_path, "r", encoding="utf-8") as f:
            return f.read()

    def load_template(self, py_file: str, template_name: Optional[str] = None) -> str:
        """
        Load HTML template relative to a Python file.

        Args:
            py_file: Pass __file__ from calling module
            template_name: Template filename (auto-detected if None -> sibling .html)
        """
        dir_path = os.path.dirname(os.path.abspath(py_file))
        if template_name is None:
            template_name = os.path.basename(py_file).replace(".py", ".html")
        html_path = os.path.join(dir_path, template_name)
        return self.load_template_file(html_path)

    def render(
        self,
        py_file: str,
        context: Optional[Dict[str, Any]] = None,
        template_name: Optional[str] = None,
    ) -> str:
        """
        Unified render function with Caspian Syntax support.
        Handles: <[[ component ]]> | <template casp-for> | <template casp-if>
        """
        # LOCAL IMPORTS to prevent circular dependency
        from casp.syntax_compiler import transpile_syntax
        from casp.components_compiler import process_imports

        ctx = context.copy() if context else {}
        template_str = self.load_template(py_file, template_name)
        
        # 1. Transpile Syntax (Convert Caspian Syntax -> Jinja)
        template_str = transpile_syntax(template_str)
        
        # 2. Extract Imports (Get mapping of "House" -> ComponentClass)
        base_dir = os.path.dirname(os.path.abspath(py_file))
        _, alias_map = process_imports(template_str, base_dir=base_dir)
        
        # 3. Define Dynamic Renderer Helper
        def render_dynamic(target, props=None, caller=None):
            props = props or {}
            
            # Resolve String Alias (e.g. "House") using local map
            comp = target
            if isinstance(target, str) and target in alias_map:
                comp = alias_map[target]
            
            # Render Component
            if callable(comp):
                if caller: props['children'] = caller()
                return comp(**props)
            
            # Fallback: Display Data
            return str(target) if target is not None else ""

        # Inject helper into context so Jinja can call it
        ctx['render_dynamic'] = render_dynamic
        
        template = self.env.from_string(template_str)
        return template.render(**ctx)

    def render_page(self, py_file: str, context: Optional[Dict[str, Any]] = None) -> str:
        """Render index.html with context."""
        return self.render(py_file, context, "index.html")

    def render_layout(self, py_file: str, context: Optional[Dict[str, Any]] = None) -> str:
        """
        Render layout.html with context.

        If context is None, returns the raw layout.html string (no Jinja render).
        """
        if context is None:
            return self.load_template(py_file, "layout.html")

        # Preserve [[ children ]] tag if caller didn't provide children
        if "children" not in context:
            context["children"] = Markup("[[ children ]]")

        return self.render(py_file, context, "layout.html")

    # ---------------------------------------------------------------------
    # Layout system (Next.js style layout discovery and wrapping)
    # ---------------------------------------------------------------------

    @dataclass
    class LayoutInfo:
        dir_path: str
        has_py: bool
        has_html: bool

        @property
        def layout_id(self) -> str:
            return hashlib.md5(self.dir_path.encode()).hexdigest()[:8]

        @property
        def py_path(self) -> str:
            return os.path.join(self.dir_path, "layout.py")

        @property
        def html_path(self) -> str:
            return os.path.join(self.dir_path, "layout.html")

    @dataclass
    class LayoutResult:
        content: str
        # Props from the tuple return (Visual Data -> [[layout.*]])
        props: Dict[str, Any] = field(default_factory=dict)
        # Metadata from the Metadata class (SEO Data -> [[metadata.*]])
        meta: Dict[str, Any] = field(default_factory=dict)

    def discover_layouts(self, route_dir: str) -> List["LayoutEngine.LayoutInfo"]:
        """Find all layouts from root to route (Next.js style)."""
        route_dir = route_dir.replace("\\", "/")
        rel_path = os.path.relpath(route_dir, "src/app").replace("\\", "/")

        layouts: List[LayoutEngine.LayoutInfo] = []

        def check_dir(dir_path: str) -> Optional[LayoutEngine.LayoutInfo]:
            dir_path = dir_path.replace("\\", "/")
            html_exists = os.path.exists(os.path.join(dir_path, "layout.html"))
            py_exists = os.path.exists(os.path.join(dir_path, "layout.py"))
            if html_exists or py_exists:
                return LayoutEngine.LayoutInfo(
                    dir_path=dir_path,
                    has_py=py_exists,
                    has_html=html_exists,
                )
            return None

        root = check_dir("src/app")
        if root:
            layouts.append(root)

        current = "src/app"
        for segment in rel_path.split("/"):
            if segment in (".", "src", "app", ""):
                continue
            current = os.path.join(current, segment).replace("\\", "/")
            found = check_dir(current)
            if found:
                layouts.append(found)

        return layouts

    def load_layout_module(self, py_path: str):
        """Dynamically load layout.py module."""
        if not os.path.exists(py_path):
            return None

        module_name = f"layout_{hashlib.md5(py_path.encode()).hexdigest()}"
        spec = importlib.util.spec_from_file_location(module_name, py_path)
        if not (spec and spec.loader):
            return None

        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def resolve_layout(
        self,
        layout: "LayoutEngine.LayoutInfo",
        context: Dict[str, Any],
    ) -> "LayoutEngine.LayoutResult":
        """
        Resolve layout content with Namespace Separation.
        - Tuple Return -> populates [[layout.*]] (Visuals)
        - Metadata Class -> populates [[metadata.*]] (SEO)
        """
        passthrough = "[[ children | safe ]]"

        if layout.has_py:
            module = self.load_layout_module(layout.py_path)

            if module and not hasattr(module, "layout"):
                raise AttributeError(
                    f"The file '{layout.py_path}' is missing the required 'def layout():' function."
                )

            if module:
                # 1. Reset Metadata Context
                _runtime_metadata.set(None)

                # 2. Execute Layout Function
                sig = inspect.signature(module.layout)
                result = module.layout(
                    context) if sig.parameters else module.layout()

                # 3. Capture SEO Metadata ([[metadata.*]])
                # Priority: Dynamic > Static
                dynamic_meta_obj = _runtime_metadata.get()
                static_meta_obj = getattr(module, 'metadata', None)
                
                meta_dict = {}
                
                def extract_meta(obj):
                    if not obj: return {}
                    d = {}
                    if obj.title: d['title'] = obj.title
                    if obj.description: d['description'] = obj.description
                    if obj.extra: d.update(obj.extra)
                    return d

                # Apply Static then Dynamic (Dynamic overrides)
                meta_dict.update(extract_meta(static_meta_obj))
                meta_dict.update(extract_meta(dynamic_meta_obj))

                # 4. Capture Layout Props ([[layout.*]])
                content = passthrough
                props_dict = {}

                if result is None:
                    if layout.has_html:
                        content = self.load_template_file(layout.html_path)
                elif isinstance(result, str):
                    content = result
                elif isinstance(result, tuple) and len(result) == 2:
                    content = result[0]
                    props_dict = result[1] or {}
                elif isinstance(result, dict):
                    if layout.has_html:
                        content = self.load_template_file(layout.html_path)
                    props_dict = result

                # 5. STRICT FILTERING
                # Remove SEO keys from Visual Props to prevent confusion
                if 'title' in props_dict:
                    del props_dict['title']
                if 'description' in props_dict:
                    del props_dict['description']

                return LayoutEngine.LayoutResult(content, props_dict, meta_dict)

        if layout.has_html:
            return LayoutEngine.LayoutResult(self.load_template_file(layout.html_path))

        return LayoutEngine.LayoutResult(passthrough)

    def get_loading_files(self) -> str:
        """Get loading indicator HTML."""
        try:
            from .loading import get_loading_files as _get_loading
            return _get_loading()
        except ImportError:
            return ""

    def _inject_meta(self, html: str, layout_id: str) -> str:
        """Inject root layout meta tag."""
        meta = f'<meta name="pp-root-layout" content="{layout_id}">'
        if "<head>" in html:
            return html.replace("<head>", f"<head>\n    {meta}")
        return html

    def render_with_nested_layouts(
        self,
        children: str,
        route_dir: str,
        page_metadata: Optional[Dict[str, Any]] = None,
        page_layout_props: Optional[Dict[str, Any]] = None,
        context_data: Optional[Dict[str, Any]] = None,
        transform_fn=None,
        component_compiler: Optional[Callable] = None,
    ) -> Tuple[str, str]:
        """
        Render page wrapped in nested layouts with strict namespace merging.
        Integrates syntax transpilation (<[[...]]>) and dynamic component resolution.
        """
        # LOCAL IMPORTS to prevent circular dependency
        from casp.syntax_compiler import transpile_syntax
        from casp.components_compiler import process_imports

        context = context_data or {}

        layouts = self.discover_layouts(route_dir)
        root_id = layouts[0].layout_id if layouts else "default"

        # === DYNAMIC RENDERER HELPER (Shared Logic) ===
        def render_dynamic_helper(target, props=None, caller=None, alias_map=None):
            props = props or {}
            alias_map = alias_map or {}
            
            comp = target
            if isinstance(target, str) and target in alias_map:
                comp = alias_map[target]
            
            if callable(comp):
                if caller: props['children'] = caller()
                return comp(**props)
            return str(target) if target is not None else ""

        # --- CASE 1: No Layouts (Direct Page Render - usually raw HTML file) ---
        if not layouts:
            # 1. Transpile
            html = transpile_syntax(children)
            # 2. Imports
            _, alias_map = process_imports(html, base_dir=route_dir)
            # 3. Context Helper
            context['render_dynamic'] = lambda t, p=None, c=None: render_dynamic_helper(t, p, c, alias_map)
            
            template = self.env.from_string(html)
            html = template.render(**context)
            
            html += self.get_loading_files()
            if transform_fn:
                html = transform_fn(html)
            return html, root_id

        # --- CASE 2: Nested Layouts ---
        current_html = children
        
        accumulated_meta = {}
        accumulated_props = {}
        resolved_layers = []

        # 1. Resolve Data
        for layout in layouts:
            res = self.resolve_layout(layout, context)
            resolved_layers.append(res)
            accumulated_meta.update(res.meta)
            accumulated_props.update(res.props)

        if page_metadata: accumulated_meta.update(page_metadata)
        if page_layout_props: accumulated_props.update(page_layout_props)

        # 2. Render Layers
        for i, layout in enumerate(reversed(layouts)):
            is_root = i == (len(layouts) - 1)
            layer_result = resolved_layers[len(layouts) - 1 - i]
            layer_content = layer_result.content

            # A. TRANSPILE (for raw HTML layouts)
            layer_content = transpile_syntax(layer_content)

            # B. EXTRACT IMPORTS (for this layout layer)
            _, alias_map = process_imports(layer_content, base_dir=layout.dir_path)
            
            # C. CONTEXT SETUP
            layer_context = context.copy()
            layer_context['render_dynamic'] = lambda t, p=None, c=None: render_dynamic_helper(t, p, c, alias_map)

            if is_root:
                current_html += self.get_loading_files()
                layer_content = self._inject_meta(layer_content, root_id)

            template = self.env.from_string(layer_content)
            current_html = template.render(
                children=Markup(current_html),
                layout=accumulated_props,
                metadata=accumulated_meta,
                **layer_context,
            )

            if component_compiler:
                current_html = component_compiler(current_html, base_dir=layout.dir_path)

        if transform_fn:
            current_html = transform_fn(current_html)

        return current_html, root_id

@dataclass
class Metadata:
    title: Optional[str] = None
    description: Optional[str] = None
    extra: Optional[Dict[str, str]] = None 

    def __post_init__(self):
        # Initialize dictionary if None
        if self.extra is None:
            self.extra = {}
            
        # 1. Dynamic Usage: Set context var (for usage inside functions)
        _runtime_metadata.set(self)
        
        # 2. Static Usage: Auto-register to module globals
        # This fixes the issue where unassigned Metadata(...) was ignored.
        try:
            # Stack: __post_init__ (0) -> __init__ (1) -> Caller (2)
            frame = inspect.currentframe()
            if frame and frame.f_back and frame.f_back.f_back:
                caller_frame = frame.f_back.f_back
                
                # Only auto-register if called at module level
                if caller_frame.f_code.co_name == '<module>':
                    # Magically set 'metadata' variable in that file
                    caller_frame.f_globals['metadata'] = self
        except Exception:
            pass


# -----------------------------------------------------------------------------
# Singleton + module-level API (single source of truth)
# -----------------------------------------------------------------------------

_engine = LayoutEngine()

# Jinja env exports
env = _engine.env
string_env = _engine.string_env

# Core exports


def load_template_file(file_path: str) -> str:
    return _engine.load_template_file(file_path)


def load_template(py_file: str, template_name: Optional[str] = None) -> str:
    return _engine.load_template(py_file, template_name)


def render(
    py_file: str,
    context: Optional[Dict[str, Any]] = None,
    template_name: Optional[str] = None,
) -> str:
    return _engine.render(py_file, context, template_name)


def render_page(py_file: str, context: Optional[Dict[str, Any]] = None) -> str:
    return _engine.render_page(py_file, context)


def render_layout(py_file: str, context: Optional[Dict[str, Any]] = None) -> str:
    return _engine.render_layout(py_file, context)


# Layout system exports
LayoutInfo = LayoutEngine.LayoutInfo
LayoutResult = LayoutEngine.LayoutResult


def discover_layouts(route_dir: str) -> List[LayoutInfo]:
    return _engine.discover_layouts(route_dir)


def load_layout_module(py_path: str):
    return _engine.load_layout_module(py_path)


def resolve_layout(layout: LayoutInfo, context: Dict[str, Any]) -> LayoutResult:
    return _engine.resolve_layout(layout, context)


def get_loading_files() -> str:
    return _engine.get_loading_files()


def render_with_nested_layouts(
    children: str,
    route_dir: str,
    page_metadata: Optional[Dict[str, Any]] = None,
    page_layout_props: Optional[Dict[str, Any]] = None,
    context_data: Optional[Dict[str, Any]] = None,
    transform_fn=None,
    component_compiler: Optional[Callable] = None,
) -> Tuple[str, str]:
    return _engine.render_with_nested_layouts(
        children=children,
        route_dir=route_dir,
        page_metadata=page_metadata,
        page_layout_props=page_layout_props,
        context_data=context_data,
        transform_fn=transform_fn,
        component_compiler=component_compiler,
    )


# Back-compat alias (if referenced elsewhere)
render_with_layouts = render_with_nested_layouts


__all__ = [
    "LayoutEngine",
    "env",
    "string_env",
    "load_template_file",
    "load_template",
    "render",
    "render_page",
    "render_layout",
    "LayoutInfo",
    "LayoutResult",
    "discover_layouts",
    "resolve_layout",
    "load_layout_module",
    "get_loading_files",
    "render_with_nested_layouts",
    "render_with_layouts",
]


from .layout import env, load_template_file
from .html_attrs import get_attributes
from .syntax_compiler import transpile_syntax
from typing import Callable, TypeVar, Dict, Optional, Any, Union
from markupsafe import Markup
import inspect
import os
import importlib
import importlib.util

R = TypeVar('R')

_component_registry: Dict[str, 'Component'] = {}


class Component:
    """Wrapper class for PulsePoint components"""
    _is_pp_component: bool = True
    source_path: Optional[str] = None

    def __init__(self, fn: Callable[..., str], source_path: Optional[str] = None):
        self.fn = fn
        self.source_path = source_path
        self.__name__ = fn.__name__
        self.__doc__ = fn.__doc__
        self.__signature__ = inspect.signature(fn)
        _component_registry[fn.__name__] = self

    def __call__(self, *args, **kwargs) -> Union[str, Markup]:
        # Standardize React-like props to Pythonic names
        if "className" in kwargs:
            kwargs["class_name"] = kwargs.pop("className")
        if "htmlFor" in kwargs:
            kwargs["html_for"] = kwargs.pop("htmlFor")

        result = self.fn(*args, **kwargs)
        if isinstance(result, str) and not isinstance(result, Markup):
            return Markup(result)

        return result


def component(fn: Callable[..., str]) -> Component:
    """Decorator to mark and register a PulsePoint component"""
    return Component(fn)


def load_component_from_path(path: str, component_name: str, base_dir: str = "") -> Component | None:
    if path == '.' or path == './':
        full_path = os.path.join(
            base_dir, f"{component_name}.py") if base_dir else f"{component_name}.py"
    elif path.startswith('./') or path.startswith('../'):
        full_path = os.path.normpath(os.path.join(
            base_dir, path)) if base_dir else os.path.normpath(path)
    else:
        full_path = path

    if full_path.endswith('.py'):
        file_path = full_path
    elif os.path.isdir(full_path):
        file_path = os.path.join(full_path, f"{component_name}.py")
    else:
        file_path = f"{full_path}.py" if not os.path.isdir(
            full_path) else os.path.join(full_path, f"{component_name}.py")

    if not os.path.exists(file_path):
        return None

    try:
        cwd = os.getcwd()
        abs_file_path = os.path.abspath(file_path)
        if abs_file_path.startswith(cwd):
            rel_path = os.path.relpath(abs_file_path, cwd)
            if not rel_path.startswith('..'):
                module_name = os.path.splitext(
                    rel_path)[0].replace(os.sep, '.')
                module = importlib.import_module(module_name)
                comp = getattr(module, component_name, None)
                if isinstance(comp, Component):
                    comp.source_path = abs_file_path
                    return comp
    except (ImportError, ValueError, AttributeError):
        pass

    try:
        unique_module_name = f"pp_component_{os.path.abspath(file_path).replace(os.sep, '_').replace('.', '_')}"
        spec = importlib.util.spec_from_file_location(
            unique_module_name, file_path)
        if spec is None or spec.loader is None:
            return None
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        comp = getattr(module, component_name, None)
        if isinstance(comp, Component):
            comp.source_path = file_path
            return comp
    except Exception as e:
        print(
            f"Error loading component {component_name} from {file_path}: {e}")
        return None

    return None


# ====
# TEMPLATE RENDERING (Jinja2 powered)
# ====

def render_html(file_path: str, **kwargs) -> str:
    """
    Load HTML template and render with Jinja2.
    """
    # Resolve file path relative to caller
    frame = inspect.currentframe()
    if frame and frame.f_back:
        caller_file = frame.f_back.f_globals.get('__file__', '')
        if caller_file:
            base_dir = os.path.dirname(caller_file)
            full_path = os.path.join(base_dir, file_path)
        else:
            full_path = file_path
    else:
        full_path = file_path

    if full_path.endswith('.py'):
        full_path = full_path[:-3] + '.html'

    content = load_template_file(full_path)
    content = transpile_syntax(content)

    # 1. EXTRACT CHILDREN
    children_content = kwargs.pop("children", "")

    # 2. MARK CHILDREN AS SAFE
    # If children comes from another component (via the fix above), it's already Markup.
    # If it's a plain string, we trust it here (React children prop behavior).
    safe_children = Markup(children_content) if children_content else ""

    # 3. GENERATE ATTRIBUTES
    # Using your NEW get_attributes from html_attrs.py which returns Markup
    props_string = get_attributes(kwargs)

    # 4. RENDER
    template = env.from_string(content)
    return template.render(props=props_string, children=safe_children, **kwargs)


def render_component(py_file: str, template_name: str, context: Optional[Dict[str, Any]] = None) -> str:
    """
    Render component template with context.
    """
    context = context or {}
    dir_path = os.path.dirname(os.path.abspath(py_file))
    html_path = os.path.join(dir_path, template_name)

    content = load_template_file(html_path)
    content = transpile_syntax(content)

    attrs = context.copy()
    children_content = attrs.pop("children", "")

    # Using the NEW get_attributes from html_attrs.py
    props_string = get_attributes(attrs)
    safe_children = Markup(children_content) if children_content else ""

    template = env.from_string(content)
    return template.render(props=props_string, children=safe_children, **context)


# ====
# EXPORTS
# ====

__all__ = [
    'Component',
    'component',
    'load_component_from_path',
    'render_html',
    'render_component',
    '_component_registry',
]



from __future__ import annotations
from .string_helpers import camel_to_kebab
from typing import Any, Dict, Optional
import html
from tailwind_merge import TailwindMerge
from markupsafe import Markup

_twm = TailwindMerge()

# Map Python-safe names to HTML attributes
ALIAS_MAP = {
    "class_name": "class",
    "className": "class",
    "html_for": "for",
    "htmlFor": "for"
}


def merge_classes(*classes: Any) -> str:
    parts: list[str] = []
    for c in classes:
        if not c:
            continue
        if isinstance(c, (list, tuple, set)):
            s = " ".join(str(x) for x in c if x)
        else:
            s = str(c).strip()
        if s:
            parts.append(s)
    return _twm.merge(*parts) if parts else ""


def _to_attr_value(v: Any) -> Optional[str]:
    # XML-strict: booleans should have explicit values
    if v is None:
        return None
    if isinstance(v, bool):
        return "true" if v else None  # omit false
    if isinstance(v, (list, tuple, set)):
        s = " ".join(str(x) for x in v if x)
        return s if s else None
    s = str(v)
    return s if s != "" else None


def get_attributes(props: Dict[str, Any], overrides: Optional[Dict[str, Any]] = None) -> Markup:
    """
    Build ONE attribute string.
    - Resolves Aliases first (className -> class).
    - Normalizes ALL keys to kebab-case (dataSlot -> data-slot).
    - Merges defaults (props) and overrides to ensure correct precedence.
    - Returns Markup() so Jinja does not double-escape the quotes.
    """
    
    final_attrs: Dict[str, str] = {}

    # Helper to process a dictionary into the final map
    def process_source(source: Dict[str, Any]):
        if not source:
            return

        for k, v in source.items():
            if not k:
                continue

            # 1. Check Alias Map first (e.g., class_name -> class)
            key_name = ALIAS_MAP.get(k, k)

            # 2. Normalize to Kebab Case ALWAYS
            # This ensures that 'dataSlot' (from compiler) overwrites 'data-slot' (from defaults)
            # regardless of whether the value has mustache syntax or not.
            final_key = camel_to_kebab(key_name)

            # 3. Process value to string
            val = _to_attr_value(v)

            if val is not None:
                final_attrs[final_key] = val

    # 1. Process defaults (props) first
    process_source(props)

    # 2. Process overrides second (this overwrites keys from step 1)
    if overrides:
        process_source(overrides)

    # 3. Build string
    parts: list[str] = []
    # Sorting keys for consistent output
    for k in sorted(final_attrs.keys()):
        v = final_attrs[k]
        parts.append(f'{k}="{html.escape(v, quote=True)}"')

    # 4. Return as Safe Markup
    return Markup(" ".join(parts))


from .string_helpers import camel_to_kebab
from .string_helpers import kebab_to_camel
from .string_helpers import has_mustache
import re
import uuid
import inspect
from typing import Dict, List, Optional
from .component_decorator import Component, load_component_from_path
import os


def convert_mustache_attrs_to_kebab_raw(html):
    def replace_attr(match):
        attr_name = match.group(1)
        equals_quote = match.group(2)
        value = match.group(3)
        end_quote = match.group(4)

        if has_mustache(value):
            new_name = camel_to_kebab(attr_name)
            return f'{new_name}{equals_quote}{value}{end_quote}'
        return match.group(0)

    pattern = r'([a-zA-Z_][a-zA-Z0-9_]*)(=["\'])([^"\']*?)(["\'])'
    return re.sub(pattern, replace_attr, html)


def extract_component_original_attrs(html, pattern):
    component_attrs_map = {}

    for match in re.finditer(pattern, html):
        component_name = match.group(1)
        tag_start = match.start()

        remaining = html[tag_start:]
        tag_end_match = re.search(r'/?>', remaining)
        if tag_end_match:
            full_tag = remaining[:tag_end_match.end()]

            attr_pattern = re.compile(
                r'([a-zA-Z_][\w\-]*)\s*=\s*(?:"[^"]*"|\'[^\']*\'|[^\s>]+)')
            attrs_original = {}
            for attr_match in attr_pattern.finditer(full_tag):
                attr_name = attr_match.group(1)
                attrs_original[attr_name.lower()] = attr_name

            placeholder = f"ppcomponent{component_name.lower()}"
            if placeholder not in component_attrs_map:
                component_attrs_map[placeholder] = []
            component_attrs_map[placeholder].append(attrs_original)

    return component_attrs_map

# ============================================================================
# IMPORT PARSING
# ============================================================================


def extract_imports(html: str) -> List[Dict]:
    """
    Extract all import statements from HTML.

    Supports:
        <!-- @import Test from "./path" -->
        <!-- @import Test from ./path -->  (unquoted)
        <!-- @import Test as MyTest from "./path" -->
        <!-- @import { Button, Input } from "./path" -->
        <!-- @import { Button as Btn, Input as Field } from "./path" -->
    """
    imports = []

    # Grouped imports: @import { A, B as C } from "path" or from path
    grouped_pattern = r'<!--\s*@import\s*\{([^}]+)\}\s*from\s*["\']?([^"\'>\s]+)["\']?\s*-->'
    grouped_re = re.compile(grouped_pattern)
    for match in grouped_re.finditer(html):
        members = match.group(1)
        path = match.group(2)

        for member in members.split(','):
            member = member.strip()
            if not member:
                continue
            if ' as ' in member:
                parts = member.split(' as ')
                original = parts[0].strip()
                alias = parts[1].strip()
            else:
                original = member
                alias = member
            imports.append({
                'original': original,
                'alias': alias,
                'path': path
            })

    # Single imports: @import Name from "path" OR @import Name as Alias from "path" (quotes optional)
    single_pattern = r'<!--\s*@import\s+(?!\{)(\w+)(?:\s+as\s+(\w+))?\s+from\s*["\']?([^"\'>\s]+)["\']?\s*-->'
    single_re = re.compile(single_pattern)
    for match in single_re.finditer(html):
        original = match.group(1)
        alias = match.group(2) or original
        path = match.group(3)
        imports.append({
            'original': original,
            'alias': alias,
            'path': path
        })

    return imports


def strip_imports(html: str) -> str:
    """Remove import comments from HTML"""
    # Remove grouped imports (quoted or unquoted)
    html = re.sub(
        r'<!--\s*@import\s*\{[^}]+\}\s*from\s*["\']?[^"\'>\s]+["\']?\s*-->\n?', '', html)
    # Remove single imports (quoted or unquoted)
    html = re.sub(
        r'<!--\s*@import\s+\w+(?:\s+as\s+\w+)?\s+from\s*["\']?[^"\'>\s]+["\']?\s*-->\n?', '', html)
    return html


def build_alias_map(imports: List[Dict], base_dir: str = "") -> Dict[str, Component]:
    """Build mapping from alias names to Component instances."""
    alias_map = {}

    for imp in imports:
        original = imp['original']
        alias = imp['alias']
        path = imp['path']

        component = load_component_from_path(path, original, base_dir)
        if component:
            alias_map[alias] = component

    return alias_map


def process_imports(html: str, base_dir: str = "") -> tuple[str, Dict[str, Component]]:
    """Process imports and return cleaned HTML with alias map."""
    imports = extract_imports(html)
    alias_map = build_alias_map(imports, base_dir)
    cleaned_html = strip_imports(html)
    return cleaned_html, alias_map


# ============================================================================
# COMPONENT HELPERS
# ============================================================================

def needs_parent_scope(attrs):
    for key, value in attrs.items():
        if key.lower().startswith('on'):
            return True
        if has_mustache(value):
            return True
    return False


def wrap_scope(html, scope):
    return f'<!-- pp-scope:{scope} -->{html}<!-- /pp-scope -->'


def accepts_children(fn):
    sig = fn.__signature__ if isinstance(
        fn, Component) else inspect.signature(fn)
    params = sig.parameters
    return 'children' in params or any(p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values())


# ============================================================================
# MAIN TRANSFORM
# ============================================================================

def transform_components(
    html: str,
    parent_scope: str = "app",
    base_dir: str = "",
    alias_map: Optional[Dict[str, Component]] = None,
    _depth: int = 0
) -> str:
    """
    Transform custom component tags. Components MUST be imported to be resolved.
    Recursively processes the output of components to handle nested imports/components.
    """
    if _depth > 50:
        return html

    # 1. Process imports for the current level
    html, local_alias_map = process_imports(html, base_dir)

    if alias_map:
        merged_map = {**alias_map, **local_alias_map}
    else:
        merged_map = local_alias_map

    # If no components are imported, we might still have standard HTML to return,
    # but we can't transform custom tags without a map.
    if not merged_map:
        return html

    # 2. Convert mustache attributes (e.g. someProp="{val}") to kebab-case (some-prop="{val}")
    html = convert_mustache_attrs_to_kebab_raw(html)

    # 3. Process components iteratively
    max_iterations = 100
    iteration = 0

    while iteration < max_iterations:
        iteration += 1

        # Find a component tag (prefer innermost by finding one without nested components)
        # This regex looks for <CapitalizedTag ... /> or <CapitalizedTag ...>...</CapitalizedTag>
        component_pattern = r'<([A-Z][a-zA-Z0-9]*)([^>]*?)(?:/>|>([\s\S]*?)</\1>)'

        match = None
        for m in re.finditer(component_pattern, html):
            tag_name = m.group(1)
            # Only process if we have an import for this tag
            if tag_name in merged_map:
                inner_content = m.group(3) or ""
                # Check if this has nested components - if so, skip for now (inner-first strategy)
                if not re.search(r'<[A-Z][a-zA-Z0-9]*[^>]*(?:/>|>)', inner_content):
                    match = m
                    break
                elif match is None:
                    match = m  # Fallback to first match if all have nested components

        if not match:
            break

        tag_name = match.group(1)
        attrs_str = match.group(2) or ""
        children = match.group(3) or ""

        component_fn = merged_map.get(tag_name)
        if not component_fn:
            # Should not happen due to the check in the loop, but safety first
            break

        # 4. Parse attributes
        props = {}
        attr_pattern = r'([a-zA-Z_][\w\-]*)\s*=\s*(?:"([^"]*)"|\'([^\']*)\'|(\S+))'
        for attr_match in re.finditer(attr_pattern, attrs_str):
            attr_name = attr_match.group(1)
            attr_value = attr_match.group(2) or attr_match.group(
                3) or attr_match.group(4) or ""
            camel_name = kebab_to_camel(attr_name)
            props[camel_name] = attr_value

        # Handle boolean attributes (no value)
        bool_attr_pattern = r'\s([a-zA-Z_][\w\-]*)(?=\s|/?>|$)(?!=)'
        for bool_match in re.finditer(bool_attr_pattern, attrs_str):
            attr_name = bool_match.group(1)
            if attr_name not in props:
                props[kebab_to_camel(attr_name)] = ""

        unique_id = f"{tag_name.lower()}_{uuid.uuid4().hex[:8]}"

        # 5. Add children if component accepts them
        if children.strip() and accepts_children(component_fn):
            props['children'] = wrap_scope(children, parent_scope)

        # 6. Render the component (get the raw string output)
        component_html = component_fn(**props)

        # 7. RECURSIVE TRANSFORM [CRITICAL FIX]
        # We must process the returned HTML to resolve imports and components
        # inside the child component (e.g., EditTodo -> Button).
        if hasattr(component_fn, 'source_path') and component_fn.source_path:
            component_dir = os.path.dirname(component_fn.source_path)
            component_html = transform_components(
                component_html,
                parent_scope=parent_scope,  # Or a new scope if you want isolation
                base_dir=component_dir,     # Use the component's directory for relative imports
                _depth=_depth + 1
            )

        # 8. Add pp-component attribute to root element of the result
        # (This marks it for PulsePoint reactivity in the DOM)
        component_html = re.sub(
            r'^(\s*<[a-zA-Z][a-zA-Z0-9]*)',
            rf'\1 pp-component="{unique_id}"',
            component_html.strip(),
            count=1
        )

        # 9. Wrap in scope if needed (for events/state binding)
        if needs_parent_scope(props):
            component_html = wrap_scope(component_html, parent_scope)

        # 10. Replace the original tag in the HTML with the transformed result
        html = html[:match.start()] + component_html + html[match.end():]

    return html



import re


def camel_to_kebab(name):
    return re.sub(r'([a-z])([A-Z])', r'\1-\2', name).lower()


def kebab_to_camel(name):
    parts = name.split('-')
    return parts[0] + ''.join(p.capitalize() for p in parts[1:])


def has_mustache(value):
    if isinstance(value, str):
        return bool(re.search(r'\{[^}]+\}', value))
    if isinstance(value, list):
        return any(has_mustache(v) for v in value)
    return False



from __future__ import annotations
import builtins as _builtins
import html
import ipaddress
import json
import mimetypes
import os
import re
import uuid as _uuid
from datetime import datetime
from decimal import Decimal, InvalidOperation, ROUND_HALF_UP
from enum import Enum
from typing import Any, Iterable, Optional, Protocol, TypeVar, Union, runtime_checkable, cast
import ulid

try:
    # Recommended
    from email_validator import validate_email as _validate_email, EmailNotValidError  # type: ignore
except ImportError:  # pragma: no cover
    _validate_email = None
    EmailNotValidError = Exception

try:
    # Optional (strtotime-like parsing)
    from dateutil.parser import parse as _dateutil_parse  # type: ignore
except ImportError:  # pragma: no cover
    _dateutil_parse = None

try:
    # Optional (real MIME sniffing)
    import magic  # type: ignore
except ImportError:  # pragma: no cover
    magic = None


# Avoid method-name collisions in type positions inside the class.
IntT = _builtins.int
FloatT = _builtins.float
StrT = _builtins.str
TEnum = TypeVar("TEnum", bound=Enum)


@runtime_checkable
class _ReadableSeekable(Protocol):
    def read(self, n: int = ...) -> bytes: ...
    def tell(self) -> int: ...
    def seek(self, offset: int, whence: int = ...) -> int: ...


class Validate:
    # -------------------------
    # Helpers
    # -------------------------

    @staticmethod
    def _php_to_py_format(fmt: str) -> str:
        """
        Minimal PHP date format -> Python strptime/strftime mapping.
        """
        mapping: dict[str, str] = {
            "Y": "%Y",
            "m": "%m",
            "d": "%d",
            "H": "%H",
            "i": "%M",
            "s": "%S",
            "u": "%f",
        }

        out: list[str] = []
        i = 0
        while i < len(fmt):
            ch = fmt[i]
            out.append(mapping.get(ch, ch))
            i += 1
        return "".join(out)

    @staticmethod
    def _is_empty_required(value: Any) -> bool:
        if value == "0":
            return False
        if value is None:
            return True
        if isinstance(value, str):
            return len(value.strip()) == 0
        if isinstance(value, (list, tuple, dict, set)):
            return len(value) == 0
        return False

    @staticmethod
    def _to_str(value: Any) -> str:
        if value is None:
            return ""
        if isinstance(value, datetime):
            return value.strftime("%Y-%m-%d %H:%M:%S")
        return str(value)

    @staticmethod
    def _parse_datetime_any(value: Any) -> Optional[datetime]:
        if value is None or value == "":
            return None
        if isinstance(value, datetime):
            return value
        s = str(value).strip()
        if not s:
            return None

        if _dateutil_parse:
            try:
                return _dateutil_parse(s)
            except Exception:
                pass

        # Fallback
        try:
            return datetime.fromisoformat(s)
        except Exception:
            pass

        for fmt in ("%Y-%m-%d", "%Y-%m-%d %H:%M:%S"):
            try:
                return datetime.strptime(s, fmt)
            except Exception:
                continue

        return None

    # -------------------------
    # String Validation
    # -------------------------

    @staticmethod
    def string(value: Any, escape_html: bool = True) -> str:
        s = Validate._to_str(value).strip()
        return html.escape(s, quote=True) if escape_html else s

    @staticmethod
    def email(value: Any) -> Optional[str]:
        if value is None:
            return None
        s = str(value).strip()
        if not s:
            return None

        if _validate_email:
            try:
                _validate_email(s)
                return s
            except EmailNotValidError:
                return None

        # Basic fallback
        return s if re.fullmatch(r"[^@\s]+@[^@\s]+\.[^@\s]+", s) else None

    @staticmethod
    def url(value: Any) -> Optional[str]:
        if value is None:
            return None
        s = str(value).strip()
        if not s:
            return None
        return s if re.match(r"^(https?://)([^/\s]+)(/.*)?$", s, re.IGNORECASE) else None

    @staticmethod
    def ip(value: Any) -> Optional[str]:
        if value is None:
            return None
        s = str(value).strip()
        try:
            ipaddress.ip_address(s)
            return s
        except Exception:
            return None

    @staticmethod
    def uuid(value: Any) -> Optional[str]:
        if value is None:
            return None
        s = str(value).strip()
        try:
            _uuid.UUID(s)
            return s
        except Exception:
            return None

    @staticmethod
    def ulid(value: Any) -> Optional[str]:
        if value is None:
            return None
        s = str(value).strip()
        if not s:
            return None
        try:
            ulid.ULID.from_str(s)
            return s
        except Exception:
            return None

    @staticmethod
    def cuid(value: Any) -> Optional[str]:
        return value if isinstance(value, str) and re.fullmatch(r"^c[0-9a-z]{24}$", value) else None

    @staticmethod
    def cuid2(value: Any) -> Optional[str]:
        if not isinstance(value, str):
            return None
        # Strict CUID2 check: lowercase a-z and 0-9 only.
        return value if re.fullmatch(r"^[a-z0-9]{20,}$", value) else None

    @staticmethod
    def nanoid(value: Any, length: int = 21) -> Optional[str]:
        if not isinstance(value, str):
            return None
        regex = f"^[0-9a-zA-Z_-]{{{length}}}$"
        return value if re.fullmatch(regex, value) else None

    @staticmethod
    def bytes(value: Any) -> Optional[str]:
        if value is None:
            return None
        s = str(value).strip()
        return s if re.fullmatch(r"^[0-9]+[kKmMgGtT]?[bB]?$", s) else None

    @staticmethod
    def xml(value: Any) -> Optional[str]:
        if value is None:
            return None
        s = str(value)
        return s if re.match(r"^<\?xml", s) else None

    # -------------------------
    # Number Validation
    # -------------------------

    @staticmethod
    def int(value: Any) -> Optional[IntT]:
        if value is None or value == "":
            return None
        if isinstance(value, bool):
            return None
        if isinstance(value, _builtins.int):
            return value

        s = str(value).strip()
        if not re.fullmatch(r"[+-]?\d+", s):
            return None
        try:
            return _builtins.int(s)
        except Exception:
            return None

    @staticmethod
    def big_int(value: Any) -> Optional[IntT]:
        return Validate.int(value)

    @staticmethod
    def float(value: Any) -> Optional[FloatT]:
        if value is None or value == "":
            return None
        if isinstance(value, bool):
            return None
        try:
            f = _builtins.float(value)
            if f != f:  # NaN
                return None
            return f
        except Exception:
            return None

    @staticmethod
    def decimal(value: Any, scale: _builtins.int = 30) -> Optional[Decimal]:
        if value is None or value == "":
            return None
        try:
            d = Decimal(str(value))
            quant = Decimal("1").scaleb(-scale)
            return d.quantize(quant, rounding=ROUND_HALF_UP)
        except (InvalidOperation, ValueError):
            return None

    # -------------------------
    # Date Validation
    # -------------------------

    @staticmethod
    def date(value: Any, fmt: str = "Y-m-d") -> Optional[str]:
        if value is None or value == "":
            return None

        if isinstance(value, datetime):
            dt = value
        else:
            s = str(value)
            py_fmt = Validate._php_to_py_format(fmt)
            try:
                dt = datetime.strptime(s, py_fmt)
            except Exception:
                return None
            if dt.strftime(py_fmt) != s:
                return None

        return dt.strftime(Validate._php_to_py_format(fmt))

    @staticmethod
    def date_time(value: Any, fmt: str = "Y-m-d H:i:s") -> Optional[str]:
        if value is None or value == "":
            return None
        dt = value if isinstance(
            value, datetime) else Validate._parse_datetime_any(value)
        if dt is None:
            return None
        return dt.strftime(Validate._php_to_py_format(fmt))

    # -------------------------
    # Boolean Validation
    # -------------------------

    @staticmethod
    def boolean(value: Any) -> Optional[bool]:
        if value is None:
            return None
        if isinstance(value, bool):
            return value
        s = str(value).strip().lower()

        true_set = {"1", "true", "on", "yes", "y"}
        false_set = {"0", "false", "off", "no", "n"}
        if s in true_set:
            return True
        if s in false_set:
            return False
        return None

    # -------------------------
    # Other Validation
    # -------------------------

    @staticmethod
    def json(value: Any) -> str:
        if isinstance(value, str):
            try:
                json.loads(value)
                return value
            except Exception as e:
                return str(e)
        return json.dumps(value, ensure_ascii=False, separators=(",", ":"))

    @staticmethod
    def is_json(value: Any) -> bool:
        if not isinstance(value, str):
            return False
        try:
            json.loads(value)
            return True
        except Exception:
            return False

    @staticmethod
    def enum(value: Any, allowed_values: Iterable[Any]) -> bool:
        return value in list(allowed_values)

    @staticmethod
    def enum_class(
        value: Union[str, IntT, TEnum, list[Union[str, IntT, TEnum]]],
        enum_cls: type[Any],
    ) -> Union[str, IntT, list[Union[str, IntT]], None]:
        if not isinstance(enum_cls, type) or not issubclass(enum_cls, Enum):
            raise ValueError(f"'{enum_cls}' is not an Enum class.")

        enum_t = cast(type[Enum], enum_cls)

        def cast_one(v: Any) -> Optional[Union[str, IntT]]:
            if isinstance(v, enum_t):
                return cast(Union[str, IntT], getattr(v, "value", None))
            for member in enum_t:
                if getattr(member, "value", None) == v:
                    return cast(Union[str, IntT], member.value)
            return None

        if isinstance(value, list):
            out: list[Union[str, IntT]] = []
            for item in value:
                c = cast_one(item)
                if c is None:
                    return None
                out.append(c)
            return out

        return cast_one(value)

    @staticmethod
    def emojis(content: str) -> str:
        emoji_map: dict[str, str] = {
            ":)": "😊", ":-)": "😊",
            ":(": "☹️", ":-(": "☹️",
            ":D": "😄", ":-D": "😄",
            ":P": "😛", ":-P": "😛",
            ";)": "😉", ";-)": "😉",
            "<3": "❤️", "</3": "💔",
            ":wave:": "👋",
            ":thumbsup:": "👍",
            ":thumbsdown:": "👎",
            ":fire:": "🔥",
            ":100:": "💯",
            ":poop:": "💩",
        }
        for k in sorted(emoji_map.keys(), key=len, reverse=True):
            content = content.replace(k, emoji_map[k])
        return content

    # -------------------------
    # Rules Engine
    # -------------------------

    @staticmethod
    def with_rules(
        value: Any,
        rules: Union[str, list[str]],
        confirmation_value: Any = None
    ) -> Union[bool, str, None]:

        if isinstance(rules, list):
            rules_array = rules
        else:
            rules_array = rules.split("|") if rules else []

        for rule in rules_array:
            if ":" in rule:
                rule_name, parameter = rule.split(":", 1)
                result = Validate.apply_rule(
                    rule_name, parameter, value, confirmation_value)
            else:
                result = Validate.apply_rule(
                    rule, None, value, confirmation_value)

            if result is not True:
                return result
        return True

    @staticmethod
    def apply_rule(rule: str, parameter: Optional[str], value: Any, confirmation_value: Any = None) -> Union[bool, str]:
        s = "" if value is None else str(value)

        if rule == "required":
            return True if not Validate._is_empty_required(value) else "This field is required."

        if rule == "min":
            n = _builtins.int(parameter or "0")
            return True if len(s) >= n else f"This field must be at least {n} characters long."

        if rule == "max":
            n = _builtins.int(parameter or "0")
            return True if len(s) <= n else f"This field must not exceed {n} characters."

        if rule == "startsWith":
            p = parameter or ""
            return True if s.startswith(p) else f"This field must start with {p}."

        if rule == "endsWith":
            p = parameter or ""
            return True if s.endswith(p) else f"This field must end with {p}."

        if rule == "confirmed":
            return True if confirmation_value == value else f"The {rule} confirmation does not match."

        if rule == "email":
            return True if Validate.email(value) else "This field must be a valid email address."

        if rule == "url":
            return True if Validate.url(value) else "This field must be a valid URL."

        if rule == "ip":
            return True if Validate.ip(value) else "This field must be a valid IP address."

        if rule == "uuid":
            return True if Validate.uuid(value) else "This field must be a valid UUID."

        if rule == "ulid":
            return True if Validate.ulid(value) else "This field must be a valid ULID."

        if rule == "cuid":
            return True if Validate.cuid(value) else "This field must be a valid CUID."

        if rule == "cuid2":
            return True if Validate.cuid2(value) else "This field must be a valid CUID2."

        if rule == "nanoid":
            return True if Validate.nanoid(value) else "This field must be a valid NanoID."

        if rule == "int":
            return True if Validate.int(value) is not None else "This field must be an integer."

        if rule == "float":
            return True if Validate.float(value) is not None else "This field must be a float."

        if rule == "boolean":
            return True if Validate.boolean(value) is not None else "This field must be a boolean."

        if rule == "in":
            opts = [x.strip()
                    for x in (parameter or "").split(",") if x.strip() != ""]
            return True if s in opts else "The selected value is invalid."

        if rule == "notIn":
            opts = [x.strip()
                    for x in (parameter or "").split(",") if x.strip() != ""]
            return True if s not in opts else "The selected value is invalid."

        if rule == "size":
            n = _builtins.int(parameter or "0")
            return True if len(s) == n else f"This field must be exactly {n} characters long."

        if rule == "between":
            parts = (parameter or "").split(",", 1)
            if len(parts) != 2:
                return "This field has an invalid between rule."
            mn, mx = _builtins.int(parts[0]), _builtins.int(parts[1])
            return True if (mn <= len(s) <= mx) else f"This field must be between {mn} and {mx} characters long."

        if rule == "date":
            fmt = parameter or "Y-m-d"
            return True if Validate.date(value, fmt) else "This field must be a valid date."

        if rule == "dateFormat":
            if not parameter:
                return "This field has an invalid dateFormat rule."
            py_fmt = Validate._php_to_py_format(parameter)
            try:
                datetime.strptime(s, py_fmt)
                return True
            except Exception:
                return f"This field must match the format {parameter}."

        if rule == "before":
            left = Validate._parse_datetime_any(value)
            right = Validate._parse_datetime_any(parameter)
            if left is None or right is None:
                return f"This field must be a date before {parameter}."
            return True if left < right else f"This field must be a date before {parameter}."

        if rule == "after":
            left = Validate._parse_datetime_any(value)
            right = Validate._parse_datetime_any(parameter)
            if left is None or right is None:
                return f"This field must be a date after {parameter}."
            return True if left > right else f"This field must be a date after {parameter}."

        if rule == "json":
            return True if Validate.is_json(value) else "This field must be a valid JSON string."

        if rule == "regex":
            if not parameter:
                return "This field format is invalid."
            try:
                return True if re.search(parameter, s) else "This field format is invalid."
            except re.error:
                return "This field format is invalid."

        if rule == "digits":
            n = _builtins.int(parameter or "0")
            return True if s.isdigit() and len(s) == n else f"This field must be {n} digits."

        if rule == "digitsBetween":
            parts = (parameter or "").split(",", 1)
            if len(parts) != 2:
                return "This field has an invalid digitsBetween rule."
            mn, mx = _builtins.int(parts[0]), _builtins.int(parts[1])
            return True if s.isdigit() and (mn <= len(s) <= mx) else f"This field must be between {mn} and {mx} digits."

        if rule == "extensions":
            exts = [x.strip()
                    for x in (parameter or "").split(",") if x.strip()]
            return True if Validate.is_extension_allowed(value, exts) else (
                "The file must have one of the following extensions: " +
                ", ".join(exts) + "."
            )

        if rule == "mimes":
            mimes_ = [x.strip()
                      for x in (parameter or "").split(",") if x.strip()]
            return True if Validate.is_mime_type_allowed(value, mimes_) else (
                "The file must be of type: " + ", ".join(mimes_) + "."
            )

        if rule == "file":
            return True if Validate.is_file(value) else "This field must be a valid file."

        return True

    # -------------------------
    # File helpers
    # -------------------------

    @staticmethod
    def is_file(file_value: Any) -> bool:
        if isinstance(file_value, str):
            return os.path.isfile(file_value)
        return isinstance(file_value, _ReadableSeekable) or hasattr(file_value, "read")

    @staticmethod
    def is_extension_allowed(file_value: Any, allowed_extensions: list[str]) -> bool:
        filename: Optional[str] = None

        if isinstance(file_value, str):
            filename = file_value
        elif hasattr(file_value, "filename"):
            maybe = getattr(file_value, "filename")
            if isinstance(maybe, str):
                filename = maybe

        if not filename:
            return False

        ext = os.path.splitext(filename)[1].lstrip(".").lower()
        allowed = {e.lower().lstrip(".") for e in allowed_extensions}
        return ext in allowed

    @staticmethod
    def is_mime_type_allowed(file_value: Any, allowed_mime_types: list[str]) -> bool:
        # Real sniffing first
        if magic is not None:
            try:
                if isinstance(file_value, str) and os.path.isfile(file_value):
                    mt = magic.from_file(file_value, mime=True)
                    return mt in allowed_mime_types

                if isinstance(file_value, _ReadableSeekable):
                    pos = file_value.tell()
                    data = file_value.read(2048)
                    file_value.seek(pos)
                    mt = magic.from_buffer(data, mime=True)
                    return mt in allowed_mime_types
            except Exception:
                pass

        # Fallback: mimetypes guess by filename
        filename: Optional[str] = None
        if isinstance(file_value, str):
            filename = file_value
        elif hasattr(file_value, "filename"):
            maybe = getattr(file_value, "filename")
            if isinstance(maybe, str):
                filename = maybe

        if not filename:
            return False

        guessed, _ = mimetypes.guess_type(filename)
        return guessed in allowed_mime_types


class Rule:
    """
    IntelliSense helper for Validation rules.
    Usage: Validate.with_rules(val, [Rule.required, Rule.min(5)])
    """
    # Simple Rules (Constants)
    REQUIRED = "required"
    EMAIL = "email"
    URL = "url"
    IP = "ip"
    UUID = "uuid"
    ULID = "ulid"
    CUID = "cuid"
    CUID2 = "cuid2"
    NANOID = "nanoid"
    INTEGER = "int"
    FLOAT = "float"
    BOOLEAN = "boolean"
    JSON = "json"
    FILE = "file"

    # Parameterized Rules (Methods)
    @staticmethod
    def min(value: int) -> str:
        return f"min:{value}"

    @staticmethod
    def max(value: int) -> str:
        return f"max:{value}"

    @staticmethod
    def size(value: int) -> str:
        return f"size:{value}"

    @staticmethod
    def starts_with(prefix: str) -> str:
        return f"startsWith:{prefix}"

    @staticmethod
    def ends_with(suffix: str) -> str:
        return f"endsWith:{suffix}"

    @staticmethod
    def confirmed() -> str:
        return "confirmed"

    @staticmethod
    def in_list(values: list[Any]) -> str:
        # joing values with comma
        val_str = ",".join(str(v) for v in values)
        return f"in:{val_str}"

    @staticmethod
    def not_in_list(values: list[Any]) -> str:
        val_str = ",".join(str(v) for v in values)
        return f"notIn:{val_str}"

    @staticmethod
    def between(min_val: int, max_val: int) -> str:
        return f"between:{min_val},{max_val}"

    @staticmethod
    def digits(count: int) -> str:
        return f"digits:{count}"

    @staticmethod
    def digits_between(min_val: int, max_val: int) -> str:
        return f"digitsBetween:{min_val},{max_val}"

    @staticmethod
    def date(fmt: str = "Y-m-d") -> str:
        return f"date:{fmt}"

    @staticmethod
    def date_format(fmt: str) -> str:
        return f"dateFormat:{fmt}"

    @staticmethod
    def before(date_str: str) -> str:
        return f"before:{date_str}"

    @staticmethod
    def after(date_str: str) -> str:
        return f"after:{date_str}"

    @staticmethod
    def regex(pattern: str) -> str:
        return f"regex:{pattern}"

    @staticmethod
    def extensions(exts: list[str]) -> str:
        return f"extensions:{','.join(exts)}"

    @staticmethod
    def mimes(types: list[str]) -> str:
        return f"mimes:{','.join(types)}"



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='caspian-utils',
    version='0.0.17',
    author='Jefferson Abraham',
    packages=find_packages(),
    install_requires=[
        "fastapi~=0.110",
        "uvicorn~=0.27",
        "python-dotenv~=1.0",
        "jinja2~=3.1",
        "beautifulsoup4~=4.12",
        "tailwind-merge~=0.1",
        "slowapi~=0.1",
        "python-multipart~=0.0.9",
        "starsessions~=1.3",
        "httpx~=0.27",
        "werkzeug~=3.0",
        "cuid2~=2.0",
        "nanoid~=2.0",
        "python-ulid~=2.7",
    ],
    description='A utility package for Caspian projects',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/TheSteelNinjaCode/caspian_utils',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.11',
)



import re
from typing import List, Tuple, Dict, Optional, Any
from casp.string_helpers import kebab_to_camel


class CaspianParser:
    """
    A robust, stack-based parser for Caspian Syntax.
    Handles nesting, attributes with special characters, and self-closing tags correctly.
    """

    def __init__(self):
        self.source = ""
        self.length = 0
        self.pos = 0
        self.output = []
        # Stack stores tuples: (block_type, context_info)
        # block_type: 'dynamic', 'for', 'if', 'else', 'elif'
        self.stack: List[Tuple[str, Any]] = []

    def feed(self, content: str):
        self.source = content
        self.length = len(content)
        self.pos = 0
        self.output = []
        self.stack = []

        while self.pos < self.length:
            # Efficiently scan for the next '<' character
            next_lt = self.source.find('<', self.pos)

            if next_lt == -1:
                # No more tags, flush remaining text
                self.output.append(self.source[self.pos:])
                self.pos = self.length
                break

            # Append text preceding the tag
            if next_lt > self.pos:
                self.output.append(self.source[self.pos:next_lt])

            self.pos = next_lt

            # Attempt to match specific Caspian tags
            # Order matters: check strictly specific syntax first
            if self._handle_dynamic_open():
                continue
            if self._handle_dynamic_close():
                continue
            if self._handle_template_open():
                continue
            if self._handle_template_close():
                continue

            # If no Caspian tag matched, treat '<' as literal text and advance
            self.output.append('<')
            self.pos += 1

    def get_result(self) -> str:
        return "".join(self.output)

    # --- Matchers & Handlers ---

    def _handle_dynamic_open(self) -> bool:
        """
        Matches <[[ expression ]] attributes ... >
        """
        # 1. Check for opening sequence <[[
        if not self.source.startswith('<[[', self.pos):
            return False

        # 2. Extract the expression inside [[ ... ]]
        # We look for the closing ]] that ends the expression part
        expr_end = self.source.find(']]', self.pos)
        if expr_end == -1:
            return False  # Malformed

        expr_start = self.pos + 3  # Skip <[[
        expression = self.source[expr_start:expr_end].strip()

        expression = sanitize_jinja_expression(expression)

        # 3. Parse attributes after the ]] but before > or />
        # Start scanning for attributes after the expression's ]]
        scan_pos = expr_end + 2

        attrs_str, end_pos, is_self_closing = self._extract_attributes_and_end(
            scan_pos)
        if end_pos == -1:
            return False  # Malformed tag structure

        # 4. Convert Attributes
        props = parse_attrs_to_dict_str(attrs_str)

        # 5. Output Logic
        if is_self_closing:
            # Case: <[[ expr ]] /> -> [[ render_dynamic(expr, props) ]]
            self.output.append(f"[[ render_dynamic({expression}, {props}) ]]")
        else:
            # Case: <[[ expr ]] > ... </[[ expr ]]>
            self.output.append(
                f"[% call(content) render_dynamic({expression}, {props}) %]")
            self.stack.append(('dynamic', expression))

        self.pos = end_pos
        return True

    def _handle_dynamic_close(self) -> bool:
        """
        Matches </[[ ... ]]>
        """
        if not self.source.startswith('</[[', self.pos):
            return False

        # Find the end of the tag
        tag_end = self.source.find('>', self.pos)
        if tag_end == -1:
            return False

        if self.stack and self.stack[-1][0] == 'dynamic':
            self.stack.pop()
            self.output.append("[% endcall %]")
        else:
            pass

        self.pos = tag_end + 1
        return True

    def _handle_template_open(self) -> bool:
        """
        Matches <template casp-action="...">
        """
        if not self.source.startswith('<template', self.pos):
            return False

        after_tag = self.pos + 9  # len('<template')
        if after_tag < self.length and self.source[after_tag].isalnum():
            return False

        attrs_str, end_pos, is_self_closing = self._extract_attributes_and_end(
            after_tag)
        if end_pos == -1:
            return False

        attrs = self._parse_attrs_to_dict(attrs_str)

        casp_action = None
        casp_value = None

        for key, value in attrs.items():
            if key.startswith('casp-'):
                casp_action = key
                casp_value = value
                break

        if not casp_action:
            return False

        if casp_value:
            casp_value = sanitize_jinja_expression(casp_value)

        if casp_action == 'casp-for':
            self.output.append(f"[% for {casp_value} %]")
            self.stack.append(('for', None))
        elif casp_action == 'casp-if':
            self.output.append(f"[% if {casp_value} %]")
            self.stack.append(('if', None))
        elif casp_action == 'casp-elif':
            self.output.append(f"[% elif {casp_value} %]")
            self.stack.append(('elif', None))
        elif casp_action == 'casp-else':
            self.output.append(f"[% else %]")
            self.stack.append(('else', None))

        self.pos = end_pos
        return True

    def _handle_template_close(self) -> bool:
        """
        Matches </template>
        """
        if not self.source.startswith('</template>', self.pos):
            return False

        if self.stack:
            block_type, _ = self.stack[-1]

            if block_type == 'for':
                self.output.append("[% endfor %]")
                self.stack.pop()
                self.pos += 11  # len('</template>')
                return True

            elif block_type == 'if':
                self.output.append("[% endif %]")
                self.stack.pop()
                self.pos += 11
                return True

            elif block_type in ('elif', 'else'):
                self.output.append("[% endif %]")
                self.stack.pop()
                self.pos += 11
                return True

        return False

    def _extract_attributes_and_end(self, start_pos: int) -> Tuple[str, int, bool]:
        cursor = start_pos
        in_quote = False
        quote_char = ''

        while cursor < self.length:
            char = self.source[cursor]

            if in_quote:
                if char == quote_char:
                    if self.source[cursor-1] != '\\':
                        in_quote = False
            else:
                if char in ('"', "'"):
                    in_quote = True
                    quote_char = char
                elif char == '>':
                    if self.source[cursor-1] == '/':
                        return self.source[start_pos:cursor-1], cursor + 1, True
                    return self.source[start_pos:cursor], cursor + 1, False

            cursor += 1

        return "", -1, False

    def _parse_attrs_to_dict(self, attrs_str: str) -> Dict[str, str]:
        attrs = {}
        pattern = r'([a-zA-Z0-9_-]+)(?:=(?:"([^"]*)"|\'([^\']*)\'|(\S+)))?'
        for match in re.finditer(pattern, attrs_str):
            key = match.group(1)
            val = match.group(2) or match.group(3) or match.group(4) or ""
            attrs[key] = val
        return attrs


def transpile_syntax(html: str) -> str:
    parser = CaspianParser()
    parser.feed(html)
    return parser.get_result()


def sanitize_jinja_expression(expr: str) -> str:
    """
    Auto-fixes common dictionary method conflicts in dot notation.
    Translates: menu.items -> menu['items']
    Ignores:    menu.items() (actual method calls)
    """
    # List of common dict methods that conflict with common variable names
    reserved = ['items', 'values', 'keys', 'update', 'pop', 'get']

    for word in reserved:
        # Match .word at a word boundary, NOT followed by (
        # Example matches: "menu.items", "menu.items "
        # Example ignores: "menu.items()"
        pattern = r'\.' + word + r'\b(?!\()'
        expr = re.sub(pattern, f"['{word}']", expr)

    return expr


def parse_attrs_to_dict_str(attrs_str: str) -> str:
    """
    Parses HTML-like attributes into a Python dict string.
    Correctly handles 'class' -> 'class_name' and boolean attributes.
    """
    if not attrs_str:
        return "{}"

    props = []
    # Matches key="val", key='val', or key=[[expr]]
    attr_pattern = r'([a-zA-Z0-9_-]+)(?:=(?:"([^"]*)"|\'([^\']*)\'|\[\[([^\]]+)\]\]))?'

    # Track which keys we've seen to handle boolean attributes vs value attributes
    seen_keys = set()

    for match in re.finditer(attr_pattern, attrs_str):
        key = match.group(1)
        val_double = match.group(2)
        val_single = match.group(3)
        val_expr = match.group(4)

        py_key = kebab_to_camel(key)

        # Handle Python reserved keywords
        if py_key == "className":
            py_key = "class_name"
        if py_key == "htmlFor":
            py_key = "html_for"
        if py_key == "class":
            py_key = "class_name"
        if py_key == "for":
            py_key = "html_for"

        seen_keys.add(key)

        if val_expr:
            # === FIX: Sanitize Props Expressions too ===
            sanitized_expr = sanitize_jinja_expression(val_expr.strip())
            props.append(f'"{py_key}": {sanitized_expr}')
        elif val_double is not None:
            props.append(f'"{py_key}": "{val_double}"')
        elif val_single is not None:
            props.append(f'"{py_key}": "{val_single}"')
        else:
            props.append(f'"{py_key}": True')

    return "{" + ", ".join(props) + "}"
